import invoke as inv

from .github import GithubIssue, sanitise_issue_title


@inv.task(aliases=("submit",))  # type: ignore
def submit_pr(c: inv.Context):
    c.run("gt submit --restack --stack --no-edit --publish --merge-when-ready")
    c.run("gt log short")


def create_branch_from_issue(c: inv.Context, selected_issue: GithubIssue):
    sanitised_title = sanitise_issue_title(selected_issue.title)
    branch_title = f"{selected_issue.number}-{sanitised_title}"
    first_commit_str = f"$'{selected_issue.title}\n\nFixes #{selected_issue.number}'"
    c.run(f"gt create {branch_title} --all -m {first_commit_str}")
    c.run(f"git commit --allow-empty -m {first_commit_str}")



from collections.abc import Sequence

from iterpy import Arr

from memium.source.document import Document
from memium.source.extractors.extractor import BasePromptExtractor
from memium.source.extractors.extractor_qa import QAPromptExtractor
from memium.source.prompts.prompt_qa import QAPrompt, QAWithoutDoc


class ReversedDefinitionExtractor(BasePromptExtractor):
    def __init__(self, qa_extractor: QAPromptExtractor) -> None:
        self.qa_extractor = qa_extractor

    def extract_prompts(self, document: Document) -> Sequence[QAPrompt]:
        qa_prompts = self.qa_extractor.extract_prompts(document)
        definition_prompts = Arr(qa_prompts).filter(lambda prompt: "Definition" in prompt.question)
        if definition_prompts.len() == 0:
            return []
        prompt = definition_prompts[0]
        answer = prompt.answer.rstrip()
        if answer[-1] in ".,!?;:":
            answer = answer[:-1]
        return [QAWithoutDoc(f"Term for '{answer}'?", document.source_path.stem, add_tags=[])]



from collections.abc import Sequence
from dataclasses import dataclass, field

import genanki

from .anki_prompt_cloze import AnkiCloze
from .anki_prompt_qa import AnkiQA


def fake_tag_factory() -> Sequence[str]:
    return ["FakeTag"]


@dataclass(frozen=True)
class FakeAnkiCloze(AnkiCloze):
    text: str = "FakeText"
    base_deck: str = "FakeBaseDeck"
    tags: Sequence[str] = field(default_factory=fake_tag_factory)
    css: str = "FakeCSS"
    uuid: int = 0
    edit_url: str = "FakeEditURL"


@dataclass(frozen=True)
class FakeAnkiQA(AnkiQA):
    base_deck: str = "FakeBaseDeck"
    tags: Sequence[str] = field(default_factory=fake_tag_factory)
    question: str = "FakeQuestion"
    answer: str = "FakeAnswer"
    css: str = "FakeCSS"
    uuid: int = 0
    edit_url: str = "FakeEditURL"
    source_title: str | None = None


from dataclasses import dataclass

import pytest


@dataclass(frozen=True)
class QAExample:
    card: AnkiQA
    deck: str


@pytest.mark.parametrize(
    ("example"),
    [
        QAExample(FakeAnkiQA(tags=["anki/deck/Subdeck"]), "FakeBaseDeck::Subdeck"),
        QAExample(
            FakeAnkiQA(tags=["anki/deck/Subdeck"], question="What are _Wikilinks_ on _Wikipedia_?"),
            "FakeBaseDeck::Subdeck::Wikilinks-Wikipedia",
        ),
        QAExample(
            FakeAnkiQA(question="On _Wikipedia_, what are _Wikilinks_?"),
            "FakeBaseDeck::Wikilinks-Wikipedia",
        ),
        QAExample(FakeAnkiQA(question="Without wikilinks?"), "FakeBaseDeck"),
    ],
)
def test_ankiqa_deck_inference(example: QAExample):
    assert example.card.deck == example.deck


def test_formatting(snapshot: genanki.Note):
    card = FakeAnkiQA(question="Q. This is a _question_?")
    note = card.to_genanki_note()
    assert snapshot == note.fields  # type: ignore



from pathlib import Path

from inline_snapshot import snapshot

from memium.source.document import DummyDocument
from memium.source.extractors.extractor_definition import ReversedDefinitionExtractor
from memium.source.extractors.extractor_qa import QAPromptExtractor


def test_extract_reversed_definition():
    doc = DummyDocument(
        content="""
Q. Definition?
A. A large African cat with a mane.""",
        source_path=Path("Lion.md"),
    )
    result = ReversedDefinitionExtractor(QAPromptExtractor("Q.", "A.")).extract_prompts(doc)

    assert len(result) == 1
    assert result[0].question == snapshot("Term for 'A large African cat with a mane'?")
    assert result[0].answer == snapshot("Lion")


def test_empty_definition_extractor():
    doc = DummyDocument(
        content="""
Q. No def?
A. Blah"""
    )
    result = ReversedDefinitionExtractor(QAPromptExtractor("Q.", "A.")).extract_prompts(doc)
    assert len(result) == 0



from pathlib import Path

from memium.destination.ankiconnect.anki_converter import AnkiPromptConverter
from memium.destination.ankiconnect.ankiconnect_gateway import ANKICONNECT_URL, AnkiConnectGateway
from memium.destination.destination import DeletePrompts, PushPrompts
from memium.destination.destination_ankiconnect import AnkiConnectDestination
from memium.destination.destination_dryrun import DryRunDestination
from memium.diff_determiner import PromptDiffDeterminer
from memium.environment import host_input_dir, in_docker
from memium.source.document_source import MarkdownDocumentSource
from memium.source.extractors.extractor_definition import ReversedDefinitionExtractor
from memium.source.extractors.extractor_qa import QAPromptExtractor
from memium.source.extractors.extractor_table import TableExtractor
from memium.source.prompt_source import DocumentPromptSource


def core(
    base_deck: str,
    input_dir: Path,
    max_deletions_per_run: int,
    dry_run: bool,
    push_all: bool = False,
):
    # Setup gateway as first step. If Anki is not running, no need to parse all the prompts.
    gateway = AnkiConnectGateway(
        ankiconnect_url=ANKICONNECT_URL,
        base_deck=base_deck,
        tmp_read_dir=host_input_dir() if in_docker() else input_dir,
        tmp_write_dir=input_dir,
        max_deletions_per_run=max_deletions_per_run,
        max_wait_seconds=3600,
    )

    dest_class = AnkiConnectDestination if not dry_run else DryRunDestination
    destination = dest_class(
        gateway=gateway,
        prompt_converter=AnkiPromptConverter(
            base_deck=base_deck,
            card_css=Path("memium/destination/ankiconnect/default_styling.css").read_text(),
        ),
    )

    # Get the inputs
    qa_extractor = QAPromptExtractor(question_prefix="Q.", answer_prefix="A.")
    source_prompts = DocumentPromptSource(
        document_ingester=MarkdownDocumentSource(directory=input_dir),
        prompt_extractors=[
            qa_extractor,
            ReversedDefinitionExtractor(qa_extractor),
            TableExtractor(),
        ],
    ).get_prompts()

    diff = PromptDiffDeterminer().sync(
        source_prompts=source_prompts, destination_prompts=destination.get_all_prompts()
    )

    delete_prompts = [command for command in diff if isinstance(command, DeletePrompts)]
    push_prompts = (
        [command for command in diff if isinstance(command, PushPrompts)]
        if not push_all
        else [PushPrompts(prompts=source_prompts)]
    )

    destination.update(commands=[*delete_prompts, *push_prompts])
